/*
 * pmap - print the address space map of a process
 *
 * Chris Rivera		<chrismrivera@gmail.com>
 * Robert Love		<rml@novell.com>
 * Werner Fink		<werner@suse.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, v2, as
 * published by the Free Software Foundation
 *
 * Copyright (C) 2003, 2005 Chris Rivera
 * Copyright (C) 2009, 2012, 2015 Werner Fink
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <sys/utsname.h>
#include <unistd.h>

#include "c.h"
#include "proc/readproc.h"
#include "proc/version.h"
#include "xalloc.h"

#define BUFFERSIZE	4096
#define OBJECTSIZE	1024

struct smap {
	const char* name;
	unsigned long data;
	unsigned long total;
	char has:1;
};

static struct smap smap[] = {
	{ "Size",		0,	0,	0},
	{ "Rss",		0,	0,	0},
	{ "Pss",		0,	0,	0},
	{ "Shared_Clean",	0,	0,	0},
	{ "Shared_Dirty",	0,	0,	0},
	{ "Private_Clean",	0,	0,	0},
	{ "Private_Dirty",	0,	0,	0},
	{ "Referenced",		0,	0,	0},
	{ "Anonymous",		0,	0,	0},
	{ "LazyFree",		0,	0,	0},
	{ "AnonHugePages",	0,	0,	0},
	{ "Shared_Hugetlb",	0,	0,	0},
	{ "Private_Hugetlb",	0,	0,	0},
	{ "Swap",		0,	0,	0},
	{ "SwapPss",		0,	0,	0},
	{ "KernelPageSize",	0,	0,	0},
	{ "MMUPageSize",	0,	0,	0},
	{ "Locked",		0,	0,	0},
	{ "VmFlags",		0,	0,	0},
	{ 0 }
};
static const size_t SmapLines = sizeof(smap)/sizeof(struct smap)-1;

enum SmapTypes {
	Size = 0,
	Rss,
	Pss,
	Shared_Clean,
	Shared_Dirty,
	Private_Clean,
	Private_Dirty,
	Referenced,
	Anonymous,
	LazyFree,
	AnonHugePages,
	Shared_Hugetlb,
	Private_Hugetlb,
	Swap,
	SwapPss,
	KernelPageSize,
	MMUPageSize,
	Locked,
	VmFlags,
};
static const size_t LastType = (size_t)(VmFlags+1);

static unsigned long long range_low;
static unsigned long long range_high = ~0ULL;
static unsigned long mapped;
static unsigned long shared;
static unsigned long private;
static FILE *smaps_fp;
static long lbits;
static const char *BLK = "";
static unsigned int WDT = 8;

static void usage(const char *cmd)
{
	fprintf(stderr, "usage: %s [options] pid [...]\n", cmd);
	fprintf(stderr, "  -d, --device         "
			"display offset and device numbers\n");
	fprintf(stderr, "  -q, --quiet          "
			"hide header and memory statistics\n");
	fprintf(stderr, "  -A, --limit=low,high "
			"limit results to the given range\n");
	fprintf(stderr, "  -V, --version        "
			"display version information\n");
	fprintf(stderr, "  -h, --help           "
			"display this help\n");
}

static char lastmap[BUFFERSIZE];
static int get_smap_data(struct smap *smap)
{
	unsigned long long data;
	int assigned, count;
	char *buff = &lastmap[0];
	struct smap* ptr;

	if (LastType != SmapLines)
		return 1;

	if (*lastmap == '\0') {
		if (!fgets(buff, BUFFERSIZE - 1, smaps_fp)) {
			if (feof(smaps_fp))
				return 0;
			return 1;
		}
	}
	assigned = sscanf(buff, "%llx-", &data);
	if (assigned != 1)
		return 1;

	count = 0;
	ptr = smap;
	while (ptr->name) {
		char *colon;

		if (!fgets(buff, BUFFERSIZE, smaps_fp)) {
			*lastmap = '\0';
			if (feof(smaps_fp))
				return 0;
			return 1;
		}
		if (*lastmap < 'A' || 'Z' < *lastmap)
			break;

		if ((colon = strchr(buff, ':'))) {
			if (*(colon+1) != 0x20)
				return 1;
			*colon = '\0';
		}
	next:
		if (strcmp(ptr->name, buff) != 0) {
			if ((ptr+1)->name) {
				count++;
				ptr++;
				goto next;
			}
			if (getenv("MALLOC_CHECK_") != NULL) {
				fprintf(stderr, "New entry in smap: %s\n", buff);
				return 1;
			}
			*lastmap = '\0';
			continue;
		}
		ptr->has = 1;

		assigned = sscanf(colon+1, "%llu",  &data);
		if (assigned == 1) {
			ptr->data = data;
			ptr->total += data;
		} else if (count != VmFlags)
			return 1;

		*lastmap = '\0';
		count++;
		ptr++;
	}

	return 0;
}

static void parse_line(pid_t pid, const char *line, int show_devices, int quiet)
{
	unsigned long long low, high, size;
	unsigned long long offset;
	unsigned long major, minor;
	int assigned;
	char read_perm, write_perm, exec_perm, access_type;
	char obj_buff[OBJECTSIZE] = "[anon]";

	assigned = sscanf(line, "%llx-%llx %c%c%c%c %llx %lx:%lx %*u %" 
		   STRINGIFY(OBJECTSIZE) "s", &low, &high, &read_perm,
		   &write_perm, &exec_perm, &access_type, &offset, &major,
		   &minor, obj_buff);

	if (assigned < 9) {
		fprintf(stderr, "failed to parse /proc/%d/maps\n", pid);
		exit(EXIT_FAILURE);
	}

	size = high - low;
	size /= 1024;
	mapped += size;

	if (smaps_fp && get_smap_data(&smap[0])) {
		fprintf(stderr, "failed to parse /proc/%d/smaps\n", pid);
		exit(1);
	}

	if (!quiet) {
		printf("START%s       SIZE ", BLK);

		if (smaps_fp) {
			printf("    RSS ");
			if (smap[Pss].has)
				printf("    PSS ");
			printf("  DIRTY ");
			if (smap[Swap].has)
				printf("   SWAP ");
			if (smap[SwapPss].has)
				printf("SWAPPSS ");
		}

		if (show_devices)
			printf("PERM OFFSET%s   DEVICE MAPPING\n", BLK);
		else
			printf("PERM MAPPING\n");
	}

	if (access_type == 's')
		shared += size;
	else if (access_type == 'p' && write_perm == 'w')
		private += size;

	if(low > range_high)
		return;
	if(high < range_low)
		return;

	printf("%0*llx %6lluK ", WDT, low, size);

	if (smaps_fp) {
		printf("%6luK ", smap[Rss].data);
		if (smap[Pss].has)
			printf("%6luK ", smap[Pss].data);
		printf("%6luK ", smap[Private_Dirty].data + smap[Shared_Dirty].data);
		if (smap[Swap].has)
			printf("%6luK ", smap[Swap].data);
		if (smap[SwapPss].has)
			printf("%6luK ", smap[SwapPss].data);
	}

	printf("%c%c%c%c ", read_perm, write_perm, exec_perm, access_type);

	if (show_devices)
		printf("%0*llx %02lx:%02lx  ", WDT, offset, major, minor);

	printf("%s\n", obj_buff);
}

int main(int argc, char *argv[])
{
	proc_t proc;
	FILE *fp;
	char path[PATH_MAX];
	char buff[BUFFERSIZE];
	const char *const name = *argv;
	int o, show_devices = 0, quiet = 0, head, ret = 0;
	struct utsname uts;

	if (uname(&uts) < 0) {
		fprintf(stderr, "error getting information about current kernel: %m\n");
		exit(EXIT_FAILURE);
	}

	if ((lbits = sysconf(_SC_LONG_BIT)) < 0) {
		fprintf(stderr, "error getting information about current kernel: %m\n");
		exit(EXIT_FAILURE);
	}
	switch (lbits) {
	case 64:
		BLK = "        ";
		WDT = 16;
		break;
	default:
		BLK = "";
		WDT = 8;
		break;
	}

	struct option longopts[] = {
		{ "help", 	0, NULL, 'h' },
		{ "version", 	0, NULL, 'V' },
		{ "quiet", 	0, NULL, 'q' },
		{ "device", 	0, NULL, 'd' },
		{ "extended", 	0, NULL, 'x' },
		{ "limit", 	0, NULL, 'A' },
		{  NULL,	0, NULL, 0 }
	};

	while ((o = getopt_long(argc, argv, "hqdxA:V", longopts, NULL)) != -1) {
		switch (o) {
		case 'V':
			printf("%s using library of %s\n", program_invocation_short_name, PACKAGE_STRING);
			return 0;
		case 'q':
			quiet = 1;
			break;
		case 'd':
			show_devices = 1;
			break;
		case 'x':
			/* ignore as we already do this */
			break;
		case 'A':
			if (!optarg || *optarg == 0) {
				usage(name);
				return 1;
			} else {
				char *low = xstrdup(optarg);
				char *high = strchr(low, ',');
				if (high) {
					*high = '\0';
					high++;
				} else
					high = low;
				if (low)
					range_low = strtoull(low, &low, 16);
				if (high)
					range_high = strtoull(high, &high, 16);
				if (low && (*low || *high)) {
					usage(name);
					return 1;
				}
			}
			break;
		case 'h':
			usage(name);
			return 0;
		default:
			usage(name);
			return 1;
		}
	}
	argc -= optind;
	argv += optind;

	if (argc <= 0) {
		usage(name);
		exit(EXIT_FAILURE);
	}

	o = argc;
	while (argc-- > 0) {
		struct smap* ptr;
		pid_t pid;
		errno = 0;
		pid = strtoul(*argv++, NULL, 10);
		if (errno) {
			perror("strtoul");
			exit(EXIT_FAILURE);
		}
		snprintf(path, PATH_MAX, "/proc/%d/maps", pid);
		fp = fopen(path, "r");
		if (!fp) {
			if (errno != ENOENT) {
				perror("fopen");
				exit(EXIT_FAILURE);
			}
			fprintf(stderr, "No process with pid %d found\n", pid);
			ret = 42;
			continue;
		}

		if (!get_proc_stats(pid, &proc)) {
			fprintf(stderr, "failed to parse /proc/%d\n", pid);
			exit(EXIT_FAILURE);
		}
		printf("%d: %s\n", pid, proc.cmd);

#if defined(DO_TEST) && (DO_TEST > 0)
		snprintf(path, PATH_MAX, "%s", "test");
#else
		snprintf(path, PATH_MAX, "/proc/%d/smaps", pid);
#endif
		smaps_fp = fopen(path, "r");

		head = quiet;
		while (fgets(buff, BUFFERSIZE - 1, fp)) {
			parse_line(pid, buff, show_devices, head);
			head = 1;
		}

		if (!quiet) {
			if (smaps_fp) {
				printf("Total:%s  ", BLK);
				printf(" %6luK", mapped);
				printf(" %6luK", smap[Rss].total);
				if (smap[Pss].has)
					printf(" %6luK", smap[Pss].total);
				printf(" %6luK", smap[Shared_Dirty].total + smap[Private_Dirty].total);
				if (smap[Swap].has)
					printf(" %6luK", smap[Swap].total);
				if (smap[SwapPss].has)
					printf(" %6luK", smap[SwapPss].total);
				printf("\n\n");
			} else 
				printf("mapped: %luK    ", mapped);

			if (smap[Referenced].has)
				printf("%luK writable-private, %luK readonly-private, %luK shared, and %luK referenced\n",
					private, mapped - private - shared, shared, smap[Referenced].total);
			else
				printf("%luK writable-private, %luK readonly-private, and %luK shared\n",
					private, mapped - private - shared, shared);
		}
		if (argc > 0)
			putchar('\n');

		ptr = smap;
		while (ptr->name) {
			ptr->data = ptr->total = 0;
			ptr->has = 0;
			ptr++;
		}
		mapped = shared = private = 0;
	}

	return ret;
}
